package edu.berkeley.qfat;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import java.util.*;
import edu.berkeley.qfat.geom.*;
import edu.berkeley.qfat.geom.Point;

// TO DO:
// - real anneal
// - solve self-intersection problem
// - get a better test model?

// FIXME: re-orient goal (how?)

public class Main extends MeshViewer {

    public static int verts = 0;

    public static final Random random = new Random();
    
    /** magnification factor */
    private static final float MAG = 1;

    public Main(StlFile stlf, Frame f) {
        super(f);

        for(int i=0; i<stlf.coordArray.length; i+=3) {
            Point p0 = new Point(stlf.coordArray[i+0].x * MAG, stlf.coordArray[i+0].y * MAG, stlf.coordArray[i+0].z * MAG);
            Point p1 = new Point(stlf.coordArray[i+1].x * MAG, stlf.coordArray[i+1].y * MAG, stlf.coordArray[i+1].z * MAG);
            Point p2 = new Point(stlf.coordArray[i+2].x * MAG, stlf.coordArray[i+2].y * MAG, stlf.coordArray[i+2].z * MAG);
            Vec n  = new Vec(stlf.normArray[i/3].x * MAG, stlf.normArray[i/3].y  * MAG, stlf.normArray[i/3].z * MAG);
            Mesh.T t  = goal.newT(p0, p1, p2, n, 0);
        }

        // rotate to align major axis -- this probably needs to be done by a human.
        goal.transform(new Matrix(new Vec(0, 0, 1), (float)(Math.PI/2)));

        float goal_width  = goal.diagonal().dot(new Vec(1, 0, 0));
        float goal_height = goal.diagonal().dot(new Vec(0, 1, 0));
        float goal_depth  = goal.diagonal().dot(new Vec(0, 0, 1));

        /*
        float width  = (float)0.6;
        float height = (float)0.08;
        float depth  = (float)0.3;
        */
        float width  = (float)0.6;
        float depth  = (float)0.08;
        float height = (float)0.3;

        float rshift = width/2;
        float lshift = -(width/2);

        //float halfup = height/2;
        float halfup = 0;

        translations = new Matrix[] {

            new Matrix(new Vec(lshift,  depth,    halfup)),
            new Matrix(new Vec(rshift,  depth,    halfup)),
            new Matrix(new Vec(lshift, -depth,    halfup)),
            new Matrix(new Vec(rshift, -depth,    halfup)),

            new Matrix(new Vec(lshift,       0,  height)),
            new Matrix(new Vec(rshift,       0,  height)),
            new Matrix(new Vec(lshift,       0, -height)),
            new Matrix(new Vec(rshift,       0, -height)),

            new Matrix(new Vec( width,           0,    0)),
            new Matrix(new Vec(-width,           0,    0)),

        };


        Point ltf = new Point(lshift,  (depth/2),  (height/2));
        Point mtf = new Point( 0.0,    (depth/2),  (height/2));
        Point rtf = new Point(rshift,  (depth/2),  (height/2));
        Point lbf = new Point(lshift, -(depth/2),  (height/2));
        Point mbf = new Point( 0.0,   -(depth/2),  (height/2));
        Point rbf = new Point(rshift, -(depth/2),  (height/2));

        Point ltc = new Point(lshift,  (depth/2), 0);
        Point mtc = new Point( 0.0,    (depth/2), 0);
        Point rtc = new Point(rshift,  (depth/2), 0);
        Point lbc = new Point(lshift, -(depth/2), 0);
        Point mbc = new Point( 0.0,   -(depth/2), 0);
        Point rbc = new Point(rshift, -(depth/2), 0);

        Point ltn = new Point(lshift,  (depth/2), -(height/2));
        Point mtn = new Point( 0.0,    (depth/2), -(height/2));
        Point rtn = new Point(rshift,  (depth/2), -(height/2));
        Point lbn = new Point(lshift, -(depth/2), -(height/2));
        Point mbn = new Point( 0.0,   -(depth/2), -(height/2));
        Point rbn = new Point(rshift, -(depth/2), -(height/2));

        
        Point[] points = new Point[] {
            ltf,
            mtf,
            rtf,
            lbf,
            mbf,
            rbf,

            ltc,
            mtc,
            rtc,
            lbc,
            mbc,
            rbc,

            ltn,
            mtn,
            rtn,
            lbn,
            mbn,
            rbn
        };


        // top
        tile.newT(ltf, mtf, mtc, null, 1);
        tile.newT(mtc, ltc, ltf, null, 1);
        tile.newT(mtf, rtf, rtc, null, 1);
        tile.newT(rtc, mtc, mtf, null, 1);

        tile.newT(ltc, mtc, mtn, null, 1);
        tile.newT(mtn, ltn, ltc, null, 1);
        tile.newT(mtc, rtc, rtn, null, 1);
        tile.newT(rtn, mtn, mtc, null, 1);

        // bottom (swap normals)
        tile.newT(mbf, lbf, mbc, null, 2);
        tile.newT(lbc, mbc, lbf, null, 2);
        tile.newT(rbf, mbf, rbc, null, 2);
        tile.newT(mbc, rbc, mbf, null, 2);

        tile.newT(mbc, lbc, mbn, null, 2);
        tile.newT(lbn, mbn, lbc, null, 2);

        tile.newT(rbc, mbc, rbn, null, 2);
        tile.newT(mbn, rbn, mbc, null, 2);


        // left
        tile.newT(ltf, ltc, lbc, null, 3);
        tile.newT(lbc, lbf, ltf, null, 3);
        tile.newT(ltc, ltn, lbn, null, 3);
        tile.newT(lbn, lbc, ltc, null, 3);

        // right (swap normals)
        tile.newT(rtc, rtf, rbc, null, 4);
        tile.newT(rbf, rbc, rtf, null, 4);
        tile.newT(rtn, rtc, rbn, null, 4);
        tile.newT(rbc, rbn, rtc, null, 4);

        // front
        tile.newT(ltn, mtn, mbn, null, 5);
        tile.newT(ltn, mbn, lbn, null, 5);
        tile.newT(mtn, rtn, rbn, null, 5);
        tile.newT(mtn, rbn, mbn, null, 5);

        // back
        tile.newT(mtf, ltf, mbf, null, 6);
        tile.newT(mbf, ltf, lbf, null, 6);
        tile.newT(rtf, mtf, rbf, null, 6);
        tile.newT(rbf, mtf, mbf, null, 6);

        for(Matrix m : translations) {
            for(Mesh.T t1 : tile) {
                for(Mesh.T t2 : tile) {
                    if (t1==t2) continue;

                    if ((t1.v1().p.times(m).minus(t2.v1().p).mag() < Mesh.EPSILON) &&
                        (t1.v2().p.times(m).minus(t2.v3().p).mag() < Mesh.EPSILON) &&
                        (t1.v3().p.times(m).minus(t2.v2().p).mag() < Mesh.EPSILON)) {
                        t1.e1().bindEdge(t2.e3());
                        t1.e2().bindEdge(t2.e2());
                        t1.e3().bindEdge(t2.e1());
                    }
                    if ((t1.v2().p.times(m).minus(t2.v1().p).mag() < Mesh.EPSILON) &&
                        (t1.v3().p.times(m).minus(t2.v3().p).mag() < Mesh.EPSILON) &&
                        (t1.v1().p.times(m).minus(t2.v2().p).mag() < Mesh.EPSILON)) {
                        t1.e2().bindEdge(t2.e3());
                        t1.e3().bindEdge(t2.e2());
                        t1.e1().bindEdge(t2.e1());
                    }
                    if ((t1.v3().p.times(m).minus(t2.v1().p).mag() < Mesh.EPSILON) &&
                        (t1.v1().p.times(m).minus(t2.v3().p).mag() < Mesh.EPSILON) &&
                        (t1.v2().p.times(m).minus(t2.v2().p).mag() < Mesh.EPSILON)) {
                        t1.e3().bindEdge(t2.e3());
                        t1.e1().bindEdge(t2.e2());
                        t1.e2().bindEdge(t2.e1());
                    }

                }
            }
        }

        //xMesh.Vert mid = lbf.getE(mbn).shatter();

        // rescale to match volume
        float factor = (float)Math.pow(tile.volume() / goal.volume(), 1.0/3.0);
        goal.transform(new Matrix(factor));

        // translate to match centroid
        goal.transform(new Matrix(tile.centroid().minus(goal.centroid())));

        //tx.e2.shatter();
        //tx.e3.shatter();


        tile.rebindPoints();

        //mid.move(new Vec((float)0,0,(float)-0.05));
        //ltn.move(new Vec((float)0,0,(float)-0.05));

        //mtf.move(new Vec(0, (float)-0.05, (float)0.05));


        System.out.println("tile volume: " + tile.volume());
        System.out.println("goal volume: " + goal.volume());

        tile.score_against = goal;
        goal.score_against = tile;
        tile.tilemesh = true;
    }

    public synchronized void breakit() {
        if (verts > 800) return;
        //while(verts < 800) {
        PriorityQueue<Mesh.E> es = new PriorityQueue<Mesh.E>();
        for(Mesh.E e : tile.edges()) es.add(e);
        for(int i=0; i<8; i++) {
            Mesh.E e = es.poll();
            verts++;
            System.out.println("shatter " + e);
            e.shatter();
        }
        tile.rebindPoints();
        //}
    }

    public synchronized void rand(float temp, Mesh.Vert p) {

        //p.reComputeError();
        p.reComputeErrorAround();
        double tile_score = tile.score();
        double goal_score = goal.score();

        Vec v = new Vec((random.nextFloat() - (float)0.5) / 200,
                        (random.nextFloat() - (float)0.5) / 200,
                        (random.nextFloat() - (float)0.5) / 200);
        /*
        Matrix inv = p.errorQuadric();
        v = new Vec(inv.d, inv.h, inv.l).norm().times(1/(float)1000);
        */

        boolean good = p.move(v);

        p.reComputeErrorAround();

        double new_tile_score = tile.score();
        double new_goal_score = goal.score();
        double tile_delta = (new_tile_score - tile_score) / tile_score;
        double goal_delta = (new_goal_score - goal_score) / goal_score;
        double delta = tile_delta + goal_delta;
        double swapProbability = Math.exp((-1 * delta) / temp);
        boolean doSwap = good && (Math.random() < swapProbability);
        //boolean doSwap = good && (tile_delta <= 0 && goal_delta <= 0);
        if (doSwap) {
            tile_score = new_tile_score;
            goal_score = new_goal_score;
            //System.out.println("score: " + tile_score + " / " + goal_score);
            hits++;
        } else {
            p.move(v.times(-1));
            misses++;
        }
    }

    float hits = 0;
    float misses = 0;
    public void anneal() throws Exception {
        int verts = 0;
        float hightemp = 10;
        float temp = hightemp;
        float last = 10;
        while(true) {
            double ratio = (hits+misses==0) ? 1 : (hits / (hits+misses));
            System.out.println("temp="+temp + " ratio="+(Math.ceil(ratio*100)));
            hits = 0;
            misses = 0;
            float gamma = 0;
            double acceptance = ratio;
            if (breaks) {
                breaks = false;
                    breakit();
                    //gamma = 1;
                    gamma = 2;
                    //temp = last * 0.8f;
                    //last = temp;
                    //temp = hightemp;
            } else
            if (acceptance > 0.96) gamma = 0.6f;
            else if (acceptance > 0.9) gamma = 0.7f;
            else if (acceptance > 0.8) gamma = 0.75f;
            else if (acceptance > 0.6) gamma = 0.8f;
            else {
                if (acceptance > 0.3) {
                    gamma = 0.9f;
                } else if (acceptance > 0.15) {
                    gamma = 0.95f;
                } else {
                    breakit();
                    gamma = 2;
                    //gamma = 1;
                    //temp = last * 0.8f;
                    //last = temp;
                    //temp = hightemp;
                }
            }
            temp = temp * gamma;

            HashSet<Mesh.Vert> hs = new HashSet<Mesh.Vert>();
            for(Mesh.Vert p : tile.vertices()) hs.add(p);
            for(int i=0; i<10; i++) {
                repaint();
                for(Mesh.Vert v : hs) rand(temp,v);
            }
            tile.rebuildPointSet();
            repaint();
            //breakit();
            repaint();
            goal.unApplyQuadricToNeighborAll();
            repaint();
            tile.recomputeAllFundamentalQuadrics();
            repaint();
            goal.applyQuadricToNeighborAll();
       }
    }

    public static void main(String[] s) throws Exception {
        StlFile stlf = new StlFile();
        stlf.load("simplefish.stl");
        Frame f = new Frame();
        Main main = new Main(stlf, f);
        f.pack();
        f.show();
        f.setSize(900, 900);
        f.doLayout();
        main.anneal();
    }

}